﻿/*****************************************************
*  
*  Copyright 2010 Akamai Technologies, Inc.  All Rights Reserved.
*  
*****************************************************
*  The contents of this file are subject to the Mozilla Public License
*  Version 1.1 (the "License"); you may not use this file except in
*  compliance with the License. You may obtain a copy of the License at
*  http://www.mozilla.org/MPL/
*   
*  Software distributed under the License is distributed on an "AS IS"
*  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
*  License for the specific language governing rights and limitations
*  under the License.
*   
*  
*  The Initial Developer of the Original Code is Akamai Technologies, Inc.
*  Portions created by Akamai Technologies, Inc. are Copyright (C) 2010 Akamai 
*  Technologies, Inc. All Rights Reserved. 
*  
*****************************************************/
package com.akamai.osmf.samples.controller
{
	import com.akamai.osmf.samples.event.DebugEvent;
	import com.akamai.osmf.samples.event.MediaLoadEvent;
	import com.akamai.osmf.utils.AkamaiMediaUtils;
	import com.akamai.osmf.utils.AkamaiStrings;
	import com.akamai.utils.URLUtils;
	
	import flash.display.Sprite;
	import flash.events.EventDispatcher;
	import flash.utils.getDefinitionByName;
	
	import mx.controls.Alert;
	
	import org.osmf.events.MediaFactoryEvent;
	import org.osmf.media.DefaultMediaFactory;
	import org.osmf.media.MediaElement;
	import org.osmf.media.MediaFactory;
	import org.osmf.media.MediaPlayer;
	import org.osmf.media.MediaResourceBase;
	import org.osmf.media.PluginInfoResource;
	import org.osmf.media.URLResource;
	import org.osmf.metadata.Metadata;
	import org.osmf.net.FMSURL;
	import org.osmf.net.StreamType;
	import org.osmf.net.StreamingURLResource;
	
	[Event(name="debug",		type="com.akamai.osmf.samples.event.DebugEvent")]
	[Event(name="loaded",		type="com.akamai.osmf.samples.event.MediaLoadEvent")]
	[Event(name="loadFailure",	type="com.akamai.osmf.samples.event.MediaLoadEvent")]
	
	public class Controller extends EventDispatcher
	{
		private static const NULL_MEDIA_ELEMENT_MESSAGE:String = "The media factory failed to create a media element. Is there a plugin required for this media?";
		
		private static const ENABLE_SSL_TRANSFER:String = "enableSSLTransfer";
		private static const BEST_EFFORT_DOWNLOAD_ON_FRAGMENT_ERROR:String = "bestEffortDownload";
		private static const PREVENT_SWITCHDOWN_ON_DROPPED_FRAMES:String = "preventSwitchDownOnDroppedFrames";
		private static const REQUEST_AMD_LONG_TOKEN:String = "requestAMDLongToken";

		// Variables with no getters/setters
		private var _mediaFactory:MediaFactory;

		// Variables with getters/setters
		private var _mediaPlayer:MediaPlayer;	// MediaPlayer is a controller class containing no UI		
		private var _netSessionMode:String;
		private var _asmEumVO:Object;
		private var _primaryToken:String;
		private var _connectionAuthParams:String;
		private var _streamAuthParams:String;
		private var _clipStart:Number;
		private var _clipEnd:Number;
		private var _startingIndex:Number;
		private var _startingBitrate:Number;
		private var _bandwidthEstimateInterval:Number;
		private var _enableLargeBuffers:Boolean = true;
		private var _fragRetryAttemptsForLostConnectivity:Number;
		private var _showAbsoluteTime:Boolean;
		private var _akamaiPluginIgnoreResource:Boolean = false;
		private var _usingF4MResourceForNonHDS:Boolean = false;
		private var _enableSSLTransfer:Boolean = false;
		private var _maxAllowedIndexAtStartup:Number;
		private var _bestEffortDownloadOnFragmentError:Boolean = false;
		private var _preventSwitchDownOnDroppedFrames:Boolean = false;
		private var _requestAMDLongToken:Boolean = false;
		
		public function Controller(mediaFactory:MediaFactory=null, mediaPlayer:MediaPlayer=null)
		{
			_mediaFactory = (mediaFactory != null) ? mediaFactory : new DefaultMediaFactory();
			_mediaPlayer = (mediaPlayer != null) ? mediaPlayer : new MediaPlayer();
			_mediaPlayer.currentTimeUpdateInterval = 100;
		}
		
		public function set primaryToken(value:String):void
		{
			_primaryToken = value;
		}

		public function set asmEumVO(value:Object):void
		{
			_asmEumVO = value;
		}

		public function get clipEnd():Number
		{
			return _clipEnd;
		}

		public function set clipEnd(value:Number):void
		{
			_clipEnd = value;
		}

		public function get clipStart():Number
		{
			return _clipStart;
		}

		public function set clipStart(value:Number):void
		{
			_clipStart = value;
		}

		public function get mediaPlayer():MediaPlayer
		{
			return _mediaPlayer;
		}
		
		public function get mediaFactory():MediaFactory
		{
			return _mediaFactory;
		}

		public function set netSessionMode(value:String):void
		{
			_netSessionMode = value;
		}
				
		public function set connectionAuthParams(value:String):void
		{
			_connectionAuthParams = value;
		}
		
		public function set streamAuthParams(value:String):void
		{
			_streamAuthParams = value;	
		}
				
		public function set useManualSwitching(value:Boolean):void
		{
			_mediaPlayer.autoDynamicStreamSwitch = !value;
		}
			
		public function set maxAllowedDynamicStreamIndex(value:Number):void
		{
			// If the value is NaN, MediaPlayer will set it to zero, so we'll
			// need to set it to the highest possible index
			if (isNaN(value) && _mediaPlayer.numDynamicStreams > 0)
			{
				value = _mediaPlayer.numDynamicStreams - 1;
			}
			
			if (!isNaN(value))
			{
				if (_mediaPlayer.playing && value > _mediaPlayer.numDynamicStreams - 1)
				{
					var msg:String = "Max allow index " + String(value) + " is out of range. The maximum value you can set is " 
						+ (_mediaPlayer.numDynamicStreams - 1); 
					Alert.show(msg);					
				}
				else
				{
					debug("setting MediaPlayer.maxAllowedDynamicStreamIndex to " + value);
					_mediaPlayer.maxAllowedDynamicStreamIndex = value;
					_maxAllowedIndexAtStartup = value;
				}
			}
		}
		
		public function set startingIndex(value:Number):void
		{
			_startingIndex = value;
		}
		
		public function set startingBitrate(value:Number):void
		{
			_startingBitrate = value;
		}
				
		public function set bandwidthEstimateInterval(value:Number):void
		{
			_bandwidthEstimateInterval = value;
		}
		
		public function set enableLargeBuffers(value:Boolean):void
		{
			_enableLargeBuffers = value;
		}
		
		public function set fragRetryAttemptsForLostConnectivity(value:Number):void
		{
			_fragRetryAttemptsForLostConnectivity = value;
		}
		
		public function set showAbsoluteTime(value:Boolean):void
		{
			_showAbsoluteTime = value;
		}
		
		public function set akamaiPluginIgnoreResource(value:Boolean):void
		{
			_akamaiPluginIgnoreResource = value;
		}
				
		public function set usingF4MResourceForNonHDS(value:Boolean):void
		{
			_usingF4MResourceForNonHDS = value;
		}
		
		public function set enableSSLTransfer(value:Boolean):void
		{
			_enableSSLTransfer = value;
		}
		
		public function set bestEffortDownloadOnFragmentError(value:Boolean):void
		{
			_bestEffortDownloadOnFragmentError = value;
		}
		
		public function set preventSwitchDownOnDroppedFrames(value:Boolean):void
		{
			_preventSwitchDownOnDroppedFrames = value;
		}
		
		public function set requestAMDLongToken(value:Boolean):void
		{
			_requestAMDLongToken = value;
		}
		
		public function loadPlugin(source:String, root:Sprite):void
		{
			debug("attempting to load plugin: "+source);

			var pluginResource:MediaResourceBase;
			
			if (source.substr(0, 4) == "http" || source.substr(0, 4) == "file")
			{
				// This is a URL, create a URLResource
				pluginResource = new URLResource(source);
			}
			else
			{
				// This is a class
				var pluginInfoRef:Class = flash.utils.getDefinitionByName(source) as Class;
				pluginResource = new PluginInfoResource(new pluginInfoRef);
			}
			
			// For player verification
			var metadata:Metadata = new Metadata();
			metadata.addValue(AkamaiStrings.AKAMAI_METADATA_KEY_DISPLAY_OBJECT, root);
			pluginResource.addMetadataValue(AkamaiStrings.AKAMAI_ADVANCED_STREAMING_PLUGIN_METADATA_NAMESPACE, metadata);
			
			loadPluginFromResource(pluginResource);
		}
		
		public function loadMedia(source:String):void
		{	
			var streamType:String = StreamType.LIVE_OR_RECORDED;
			
			// StreamType.LIVE_OR_RECORDED is fine for most stream types, except for live RTMP. We don't want
			// the seek trait to get added since Akamai doesn't support DVR for RTMP live so let's look for
			// an RTMP live stream
			var url:FMSURL = new FMSURL(source);
			if (url.protocol.search(/^rtmp$|rtmp[tse]$|rtmpte$/i) != -1 && 
				(url.streamName.search(/@/) != -1) || (url.streamName.search("live") != -1))
			{
				streamType = StreamType.LIVE;
			}
						
			if (source.indexOf(ENABLE_SSL_TRANSFER) != -1)
			{
				enableSSLTransfer = (URLUtils.queryArgFromString(source, ENABLE_SSL_TRANSFER) == "true");					
			}
			
			if (source.indexOf(BEST_EFFORT_DOWNLOAD_ON_FRAGMENT_ERROR) != -1)
			{
				bestEffortDownloadOnFragmentError = (URLUtils.queryArgFromString(source, BEST_EFFORT_DOWNLOAD_ON_FRAGMENT_ERROR) == "true");					
			}
			
			if (source.indexOf(PREVENT_SWITCHDOWN_ON_DROPPED_FRAMES) != -1)
			{
				preventSwitchDownOnDroppedFrames = (URLUtils.queryArgFromString(source, PREVENT_SWITCHDOWN_ON_DROPPED_FRAMES) == "true");
			}
			
			
			requestAMDLongToken = (source.indexOf(REQUEST_AMD_LONG_TOKEN) != -1) ? URLUtils.queryArgFromString(source, REQUEST_AMD_LONG_TOKEN) == "true" : false;
			
			var resource:URLResource = new StreamingURLResource(source, streamType, getStartTime(), getEndTime());
			addMetadata(resource);
			
			var mediaElement:MediaElement = mediaFactory.createMediaElement(resource);
			if (mediaElement != null)
			{
				dispatchEvent(new MediaLoadEvent(MediaLoadEvent.LOADED, false, false, mediaElement));
			}
			else
			{
				dispatchEvent(new MediaLoadEvent(MediaLoadEvent.LOAD_FAILURE, false, false, NULL_MEDIA_ELEMENT_MESSAGE));
				dispatchEvent(new DebugEvent(DebugEvent.DEBUG, false, false, "mediaFactory.createMediaElement() returned null for source: "+source));
			}
		}						
		
		private function addMetadata(resource:URLResource):void
		{
			var metadata:Metadata = resource.getMetadataValue(AkamaiStrings.AKAMAI_ADVANCED_STREAMING_PLUGIN_METADATA_NAMESPACE)
												as Metadata;
			if (metadata == null)
			{
				metadata = new Metadata();		
			}
			
			if (_netSessionMode != null && _netSessionMode.length > 0)
			{
				metadata.addValue(AkamaiStrings.AKAMAI_METADATA_KEY_NETSESSION_MODE, _netSessionMode);
			}
			
			if (_asmEumVO != null)
			{				
				metadata.addValue(AkamaiStrings.AKAMAI_METATDATA_KEY_ENABLE_ALTERNATE_SERVER_MAPPING, _asmEumVO.asm);
				metadata.addValue(AkamaiStrings.AKAMAI_METATDATA_KEY_ENABLE_END_USER_MAPPING, _asmEumVO.eum);
			}
			
			if (_connectionAuthParams != null && _connectionAuthParams.length > 0)
			{
				metadata.addValue(AkamaiStrings.AKAMAI_METADATA_KEY_CONNECT_AUTH_PARAMS, _connectionAuthParams);
			}
			
			if (_streamAuthParams != null && _streamAuthParams.length > 0)
			{
				metadata.addValue(AkamaiStrings.AKAMAI_METADATA_KEY_STREAM_AUTH_PARAMS, _streamAuthParams);
			}
			
			if (_primaryToken != null && _primaryToken.length > 0)
			{
				metadata.addValue(AkamaiStrings.AKAMAI_METADATA_KEY_PRIMARY_TOKEN, _primaryToken);
			}
			
			if (!isNaN(_startingIndex))
			{
				metadata.addValue(AkamaiStrings.AKAMAI_METADATA_KEY_MBR_STARTING_INDEX, _startingIndex);
			}
			
			if (!isNaN(_startingBitrate))
			{
				metadata.addValue(AkamaiStrings.AKAMAI_METADATA_KEY_MBR_STARTING_BITRATE, _startingBitrate);
			}
			
			if (!isNaN(_fragRetryAttemptsForLostConnectivity))
			{
				metadata.addValue(AkamaiStrings.AKAMAI_METADATA_KEY_FRAG_RETRY_ATTEMPTS, _fragRetryAttemptsForLostConnectivity);
			}
						
			if (!isNaN(_bandwidthEstimateInterval) && _bandwidthEstimateInterval > 0)
			{
				var setBandwidthEstimationEnabled:Object = new Object();
				setBandwidthEstimationEnabled.enabled = true;
				setBandwidthEstimationEnabled.bandwidthEstimationPeriodInSeconds = _bandwidthEstimateInterval
				metadata.addValue(AkamaiStrings.AKAMAI_METADATA_KEY_SET_BANDWIDTH_ESTIMATION_ENABLED, setBandwidthEstimationEnabled);
			}
			
			if (_enableSSLTransfer)
			{
				metadata.addValue(AkamaiStrings.AKAMAI_METADATA_KEY_ENABLE_SSL_TRANSFER, _enableSSLTransfer);
			}
			
			if (!_bestEffortDownloadOnFragmentError)
			{
				metadata.addValue(AkamaiStrings.AKAMAI_METADATA_KEY_ENABLE_BEST_EFFORT_DOWNLOAD_ON_FRAGMENT_ERROR, _bestEffortDownloadOnFragmentError);
			}
			
			// Jira HDCore-133
			metadata.addValue(AkamaiStrings.AKAMAI_METADATA_KEY_HDN2_PREVENT_SWITCHDOWN_ON_DROPPED_FRAMES, _preventSwitchDownOnDroppedFrames);
			
			//Jira HDCORE-141
			if (_requestAMDLongToken)
			{
				metadata.addValue(AkamaiStrings.AKAMAI_METADATA_KEY_REQUEST_AMD_LONG_TOKEN, resource.url);
			}
			else
			{
				metadata.removeValue(AkamaiStrings.AKAMAI_METADATA_KEY_REQUEST_AMD_LONG_TOKEN);
			}
									
			metadata.addValue(AkamaiStrings.AKAMAI_METADATA_KEY_ENABLE_LARGE_BUFFERS_FOR_LONG_FORM_CONTENT, _enableLargeBuffers);
			
			resource.addMetadataValue(AkamaiStrings.AKAMAI_ADVANCED_STREAMING_PLUGIN_METADATA_NAMESPACE, metadata);
			
			if (_akamaiPluginIgnoreResource)
			{
				AkamaiMediaUtils.addAkamaiMediaTypeToResource(resource, AkamaiStrings.AKAMAI_MEDIA_TYPE_IGNORE,null);
			}
			else if (_usingF4MResourceForNonHDS)
			{
				AkamaiMediaUtils.addAkamaiMediaTypeToResource(resource, AkamaiStrings.AKAMAI_MEDIA_TYPE_UNKNOWN, null);
			}

			if (!isNaN(_maxAllowedIndexAtStartup))
			{
				metadata.addValue(AkamaiStrings.AKAMAI_METADATA_KEY_MAX_ALLOWED_INDEX, _maxAllowedIndexAtStartup);
			}
			
		}
		
		private function loadPluginFromResource(pluginResource:MediaResourceBase):void
		{
			setupListeners();
			mediaFactory.loadPlugin(pluginResource);
			
			function setupListeners(add:Boolean=true):void
			{
				if (add)
				{
					mediaFactory.addEventListener(MediaFactoryEvent.PLUGIN_LOAD, onPluginLoad);
					mediaFactory.addEventListener(MediaFactoryEvent.PLUGIN_LOAD_ERROR, onPluginLoadError);					
				}
				else
				{
					mediaFactory.removeEventListener(MediaFactoryEvent.PLUGIN_LOAD, onPluginLoad);
					mediaFactory.removeEventListener(MediaFactoryEvent.PLUGIN_LOAD_ERROR, onPluginLoadError);
				}
			}
			
			function onPluginLoad(event:MediaFactoryEvent):void
			{
				debug("plugin loaded successfully.");
				setupListeners(false);
			}
			
			function onPluginLoadError(event:MediaFactoryEvent):void
			{
				debug("plugin failed to load.");
				setupListeners(false);
			}
		}
		
		private function getStartTime():Number
		{
			var t:Number = _clipStart;
			
			// Convert to unix time for live..
			if (_showAbsoluteTime && !isNaN(t))
			{
				t = getUnixTime() - _clipStart;	
			}
			
			return t;
		}
		
		private function getUnixTime():Number
		{
			return Math.floor(new Date().getTime() / 1000);
		}

		private function getEndTime():Number
		{
			// The OSMF NetStreamUtils class will take the clip end time and subtract
			// from the clip start time and use that for a length value for the
			// NetStream.play() method. But if we're using UTC time for start time
			// we need adjust this value so it will get passed into the Akamai
			// NetStream class correctly;
			
			var t:Number = _clipEnd;
			
			if (_showAbsoluteTime)
			{
				t = getStartTime() + _clipEnd;
			}
			
			return t;
		}
				
		private function debug(msg:String):void
		{
			dispatchEvent(new DebugEvent(DebugEvent.DEBUG, false, false, msg));
		}
	}
}
